#include "pch.h"
#include "SerialControl.h"


SerialControl::SerialControl()
{
	m_bPortReady = FALSE;
	lastdata = NULL; //2017/8/23
	waithandle = NULL;
	m_hCom = NULL;
}


SerialControl::~SerialControl()
{
	close();
	if (lastdata != NULL) //2017/8/23
		delete(lastdata);
	if (waithandle){
		CloseHandle(waithandle);
		waithandle = NULL;
	}
}


BOOL SerialControl::init(CString port)
{
	//http://www.ontrak.net/mfc.htm
	m_bPortReady = FALSE;
	m_sComPort = L"\\\\.\\"+port;
	m_hCom = CreateFile(m_sComPort,
		GENERIC_READ | GENERIC_WRITE,
		0, // exclusive access
		NULL, // no security
		OPEN_EXISTING,
		FILE_FLAG_OVERLAPPED,
		NULL); // null template

	if (m_hCom == INVALID_HANDLE_VALUE)
		return false;

	m_bPortReady = SetupComm(m_hCom, 512, 512); // set buffer sizes
	PurgeComm(m_hCom, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR);

	m_bPortReady = GetCommState(m_hCom, &m_dcb);
	m_dcb.BaudRate = 115200; // 9600;
	m_dcb.ByteSize = 8;
	m_dcb.Parity = NOPARITY;
	m_dcb.StopBits = ONESTOPBIT;
	m_dcb.fAbortOnError = TRUE;

	//m_dcb.fDtrControl = 0;
	//m_dcb.fRtsControl = 0;
	//m_dcb.EofChar = 0x1a;
	//m_dcb.EvtChar = 0x1a;
	//m_dcb.fNull = TRUE;

	m_bPortReady = SetCommState(m_hCom, &m_dcb);

	m_bPortReady = GetCommTimeouts(m_hCom, &m_CommTimeouts);

	m_CommTimeouts.ReadIntervalTimeout = 50;
	m_CommTimeouts.ReadTotalTimeoutConstant = 50;
	m_CommTimeouts.ReadTotalTimeoutMultiplier = 10;
	m_CommTimeouts.WriteTotalTimeoutConstant = 50;
	m_CommTimeouts.WriteTotalTimeoutMultiplier = 10;

	m_bPortReady = SetCommTimeouts(m_hCom, &m_CommTimeouts);

	return m_bPortReady;
}

void SerialControl::stopreceive(){
	if (waithandle != NULL)
		SetEvent(waithandle);
}

void SerialControl::close()
{
	if (waithandle != NULL)
		SetEvent(waithandle);
	if (m_hCom != NULL)
		CloseHandle(m_hCom);
	m_hCom = NULL;
}

BOOL SerialControl::resendData(){ //2017/8/23
	return sendData(lastdata, lastsize, FALSE);
}


BOOL SerialControl::sendData(BYTE* data,int size, BOOL copylast)
{
	cmutex.Lock();
	if (!m_bPortReady)
		return FALSE;

	//http://www.ys-labo.com/BCB/2007/070512%20RS232C%20zenpan.html

	if (copylast){ //2017/8/23
		if (lastdata != NULL) delete lastdata;
		lastdata = new BYTE[size];
		memcpy(lastdata, data, size);
		lastsize = size;
	}

	OVERLAPPED osWrite = { 0 };
	COMSTAT Comstat; // ʐMfoCX\
	DWORD dwErrorMask; // G[R[h󂯎ϐ
	ClearCommError(m_hCom,&dwErrorMask,&Comstat);

	osWrite.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

	/*CFile cfile;
	cfile.Open(L"C:\\temp\\debug.txt", CFile::modeWrite | CFile::modeCreate | CFile::typeBinary | CFile::modeNoTruncate);
	cfile.SeekToEnd();
	cfile.Write(data, size);
	cfile.Close();*/

	int ttlsent = 0;
	DWORD dwRet;
	bWriteRC = WriteFile(m_hCom, data, size, &iBytesWritten, &osWrite);
	if (!bWriteRC){
		dwRet = GetLastError();
		if (dwRet == ERROR_IO_PENDING)
			GetOverlappedResult(m_hCom, &osWrite, &iBytesWritten, TRUE);
		else{
			cmutex.Unlock();
			return FALSE;
		}
	}
	ttlsent += iBytesWritten;
	while (ttlsent < size){
		bWriteRC = WriteFile(m_hCom, &data[ttlsent], size - ttlsent, &iBytesWritten, &osWrite);
		if (!bWriteRC){
			dwRet = GetLastError();
			if (dwRet == ERROR_IO_PENDING)
				GetOverlappedResult(m_hCom, &osWrite, &iBytesWritten, TRUE);
			else{
				cmutex.Unlock();
				return FALSE;
			}
		}
		ttlsent += iBytesWritten;
	}
	cmutex.Unlock();
	return TRUE;
}

BOOL SerialControl::isopen(){
	return m_bPortReady;
}

//http://www.codeproject.com/Articles/2682/Serial-Communication-in-Windows
//https://code.google.com/p/binaryeditorbz/source/detail?r=358fe4a4e81d3b1632f1f750a452e0b06cd892d2
int SerialControl::ReceiveData(int timeout){
	if (!m_bPortReady)
		return 0;

	if (waithandle == NULL)
		waithandle = CreateEvent(0, true, 0, 0);
	else
		ResetEvent(waithandle);
	DWORD dwEventMask;
	DWORD dwBytesRead;

	OVERLAPPED ov,ovr;
	memset(&ov, 0, sizeof(ov));
	ov.hEvent = CreateEvent(0, true, 0, 0);
	memset(&ovr, 0, sizeof(ovr));
	ovr.hEvent = CreateEvent(0, true, 0, 0);
	HANDLE arHandles[2];
	arHandles[0] = waithandle;
	
	SetCommMask(m_hCom, EV_RXCHAR);

	DWORD dwWait;
	DWORD dwRet;
	DWORD ttlread = 0;
	BOOL abRet = ::WaitCommEvent(m_hCom, &dwEventMask, &ov);
	arHandles[1] = ov.hEvent;

	memset(recvBuffer, 0, 512);
	while (1) {
		dwWait = WaitForMultipleObjects(2, arHandles, FALSE, timeout);
		switch (dwWait)
		{
		case WAIT_OBJECT_0:
			CloseHandle(ov.hEvent);
			CloseHandle(ovr.hEvent);
			return 0;
		case WAIT_OBJECT_0 + 1:
		{
			if (ttlread < sizeof(recvBuffer)) {
				ResetEvent(ovr.hEvent);
				if (!ReadFile(m_hCom, &recvBuffer[ttlread], 511 - ttlread, &dwBytesRead, &ovr)) {
					dwRet = GetLastError();
					if (dwRet == ERROR_IO_PENDING)
						GetOverlappedResult(m_hCom, &ovr, &dwBytesRead, TRUE);
					else {
						CloseHandle(ov.hEvent);
						CloseHandle(ovr.hEvent);
						return 0;
					}
				}
			}
			else
				break;
			ttlread += dwBytesRead;// strlen(recvBuffer);
			recvSize = ttlread;

			if (ttlread > 0) { //strlen(recvBuffer) > 0) {
				CloseHandle(ov.hEvent);
				CloseHandle(ovr.hEvent);
				return 1;
			}
			else {
				CloseHandle(ov.hEvent);
				CloseHandle(ovr.hEvent);
				return 2;
			}
		}
		}//switch
	}
	CloseHandle(ov.hEvent);
	CloseHandle(ovr.hEvent);
	return 0;
}

void SerialControl::clearBuffer(){
	//https://docs.microsoft.com/ja-jp/windows/desktop/api/winbase/nf-winbase-purgecomm
	if (m_hCom != NULL)
		PurgeComm(m_hCom, PURGE_RXCLEAR | PURGE_RXABORT);// | PURGE_TXCLEAR);
}